/*
 *  AudioCueObjViewImpl.scala
 *  (Mellite)
 *
 *  Copyright (c) 2012-2022 Hanns Holger Rutz. All rights reserved.
 *
 *  This software is published under the GNU Affero General Public License v3+
 *
 *
 *  For further information, please contact Hanns Holger Rutz at
 *  contact@sciss.de
 */

package de.sciss.mellite.impl.objview

import de.sciss.asyncfile.Ops.URIOps
import de.sciss.audiofile.AudioFile
import de.sciss.desktop
import de.sciss.desktop.FileDialog
import de.sciss.equal.Implicits._
import de.sciss.file._
import de.sciss.lucre.swing.LucreSwing.deferTx
import de.sciss.lucre.swing.Window
import de.sciss.lucre.swing.graph.AudioFileIn
import de.sciss.lucre.synth.Txn
import de.sciss.lucre.{Artifact, ArtifactLocation, Obj, Source, Txn => LTxn}
import de.sciss.mellite.AudioCueObjView.{Config, LocationConfig, MakeResult, SingleConfig}
import de.sciss.mellite.impl.ObjViewCmdLineParser
import de.sciss.mellite.impl.objview.ObjViewImpl.{GainArg, TimeArg}
import de.sciss.mellite.{ActionArtifactLocation, AudioCueFrame, AudioCueObjView, DragAndDrop, GUI, MessageException, ObjListView, ObjView, ObjectActions, UniverseHandler, WorkspaceCache}
import de.sciss.proc.{AudioCue, TimeRef, Universe}
import de.sciss.processor.Processor.Aborted

import java.awt.datatransfer.Transferable
import java.net.URI
import scala.concurrent.Future
import scala.swing.{Component, Label}
import scala.util.{Failure, Success, Try}

object AudioCueObjViewImpl extends AudioCueObjView.Companion {
  def install(): Unit =
    AudioCueObjView.peer = this

  override def initMakeDialog[T <: Txn[T]](window: Option[desktop.Window])
                                          (implicit universe: Universe[T]): MakeResult[T] = {
    import de.sciss.lucre.synth.Executor.executionContext
    import universe.{cursor, workspace}
    val dirIn = cursor.step { implicit tx =>
      dirCache.get[T]()
    }
    val dlg = FileDialog.open(init = dirIn, title = "Add Audio Files")
    dlg.setFilter(f => f.isDirectory || Try(AudioFile.identify(f).isDefined).getOrElse(false))
    dlg.multiple = true
    val dlgOk = GUI.optionToAbortedFut(dlg.show(window))

    val res = dlgOk.flatMap { f0 =>
      val dirOutOpt = f0.parentOption
      dirOutOpt.foreach { dirOut =>
        cursor.step { implicit tx =>
          dirCache.set[T](dirOut)
        }
      }
      val fs = dlg.files
      if (fs.isEmpty) Future.failed(Aborted())
      else {
        // @tailrec
        def loop(rem: List[File], locSeq: List[LocationConfig[T]], res: Config[T]): MakeResult[T] =
          rem match {
            case head :: tail =>
              Future { AudioFile.readSpec(head) } .flatMap { spec =>
                val locExist: Option[LocationConfig[T]] = locSeq.find { case (_, dir) =>
                  val locOk = Try(Artifact.Value.relativize(parent = dir, sub = head.toURI)).isSuccess
                  locOk
                }

                val locOpt = locExist.orElse(
                  ActionArtifactLocation.query[T](file = head.toURI, window = window)(implicit tx => universe.workspace.root)
                )

                locOpt match {
                  case Some(loc) =>
                    loop(rem = tail, locSeq = loc :: locSeq, res =
                      SingleConfig(name = head.base, file = head, spec = spec, location = loc) :: res)

                  case None =>
                    Future.failed(Aborted())
                }
              }

            case Nil => Future.successful(res.reverse)
          }

        loop(fs, Nil, Nil)
      }
    }
    res
  }

  private lazy val dirCache = WorkspaceCache[File]()

  def mkListView[T <: Txn[T]](obj: AudioCue.Obj[T])(implicit tx: T): AudioCueObjView[T] with ObjListView[T] = {
    val value = obj.value
    new Impl(tx.newHandle(obj), value).init(obj)
  }

  private final case class Config2(name: Option[String] = None, file: File = null,
                                   location: Option[(String, File)] = None, offset: TimeArg = TimeArg.Frames(0L),
                                   gain: Double = 1.0,
                                   const: Boolean = false)

  override def initMakeCmdLine[T <: Txn[T]](args: List[String])(implicit universe: Universe[T]): MakeResult[T] = {
    // we do not support any number of files except one
    val default = Config2()
    object p extends ObjViewCmdLineParser[Config2](AudioCueObjView, args) {
      val location: Opt[File] = opt(
        descr = "Artifact's base location (directory). If absent, artifact's direct parent is used."
      )
      validateFileIsDirectory(location)

      val const: Opt[Boolean] = opt(descr = "Make constant instead of variable")

      val gain: Opt[GainArg] = opt(default = Some(GainArg(default.gain)),
        descr = s"Gain (linear, -3dB, ...; default: ${default.gain})"
      )
      val offset: Opt[TimeArg] = opt(default = Some(default.offset),
        descr = s"Offset into the file (frames, 1.3s, ...; default: ${default.offset})"
      )

      val file: Opt[File] = trailArg[File](descr = "File")

      validate(file) { f: File =>
        val tr = Try(AudioFile.identify(f))
        tr match {
          case Success(Some(_)) => Right(())
          case Success(None)    => Left(s"Cannot identify audio file: $f")
          case Failure(ex)      => Left(s"Cannot identify audio file: ${ex.getMessage}")
        }
      }
    }

    def resolveLoc(f: URI, opt: Option[(String, URI)]): Future[LocationConfig[T]] = opt match {
      case Some((nm, base)) => Future.successful((Right(nm), base))
      case None =>
        f./*absolute.*/parentOption match {
          case Some(parent) =>
            val vec = ActionArtifactLocation.find[T](file = f)(implicit tx => universe.workspace.root)
            val exist = vec.collectFirst {
              case l if l.value._2 === parent => (Left(l.value._1), l.value._2): LocationConfig[T]
            }
            val res = exist match {
              case Some(res0) => res0
              case None       => (Right(parent.name), parent)
            }
            Future.successful(res)

          case None => Future.failed(MessageException(s"No parent directory to '$f'"))
        }
    }

    import de.sciss.lucre.synth.Executor.executionContext

    for {
      c     <- p.parseFut(Future.successful(Config2(name = p.nameOption.orElse(p.location.toOption.map(_.name)), file = p.file(),
        offset = p.offset(), gain = p.gain().linear, const = p.const())))
      loc   <- resolveLoc(c.file.toURI, c.location.map(tup => (tup._1, tup._2.toURI)))
      spec  <- AudioFile.readSpecAsync(c.file.toURI)
    } yield {
      val c1 = SingleConfig(name = c.name.getOrElse(c.file.base), file = c.file, spec = spec, location = loc,
        offset = c.offset match {
          case TimeArg.Frames(n) => (n * TimeRef.SampleRate / spec.sampleRate + 0.5).toLong
          case other => other.frames()
        }, gain = c.gain, const = c.const)
      c1 :: Nil
    }
  }

  def makeObj[T <: Txn[T]](config: Config[T])(implicit tx: T): List[Obj[T]] = {
    var locMade = Map.empty[(String, URI), ArtifactLocation[T]]
    var res     = List.empty[Obj[T]]

    config.foreach { c =>
      val loc = c.location match {
        case (Left(source), _) => source()
        case (Right(name), directory) =>
          locMade.get((name, directory)) match {
            case Some(loc0) => loc0
            case None =>
              val loc0 = ActionArtifactLocation.create(name = name, directory = directory)
              locMade += (name, directory) -> loc0
              res ::= loc0
              loc0
          }
      }
      val audioObj = ObjectActions.mkAudioFile(loc, c.file.toURI, c.spec, offset = c.offset, gain = c.gain, const = c.const,
        name = Some(c.name))
      res ::= audioObj
    }
    res.reverse
  }

  trait Basic[T <: Txn[T]] extends ObjViewImpl.Impl[T] with AudioCueObjView[T] {
    final override def obj(implicit tx: T): AudioCue.Obj[T] = objH()

    final def factory: ObjView.Factory = AudioCueObjView

    def isViewable = true

    override def createTransferable(): Option[Transferable] = {
      val fileOpt = Try(new File(value.artifact)).toOption
      fileOpt.map { f =>
        DragAndDrop.Transferable.files(f)
      }
    }

    override def openView(parent: Option[Window[T]])(implicit tx: T,
                                                     handler: UniverseHandler[T]): Option[Window[T]] = {
      val frame = AudioCueFrame(obj)
      Some(frame)
    }
  }

  private final class Impl[T <: Txn[T]](val objH: Source[T, AudioCue.Obj[T]],
                                        var value: AudioCue)
    extends ObjListViewImpl.NonEditable[T]
      with Basic[T] {

    type E[~ <: LTxn[~]] = AudioCue.Obj[~]

    def init(obj: AudioCue.Obj[T])(implicit tx: T): this.type = {
      initAttrs(obj)
      addDisposable(obj.changed.react { implicit tx =>upd =>
        deferTx {
          value = upd.now
        }
        fire(ObjView.Repaint(this))
      })
      this
    }

    def configureListCellRenderer(label: Label): Component = {
      val txt     = AudioFileIn.specToString(value.spec)
      // XXX TODO: add offset and gain information if they are non-default
      label.text  = txt
      label
    }
  }
}
