/*
 *  ObjViewCmdLineParser.scala
 *  (Mellite)
 *
 *  Copyright (c) 2012-2022 Hanns Holger Rutz. All rights reserved.
 *
 *  This software is published under the GNU Affero General Public License v3+
 *
 *
 *  For further information, please contact Hanns Holger Rutz at
 *  contact@sciss.de
 */

package de.sciss.mellite.impl

import de.sciss.lucre.Txn
import de.sciss.mellite.impl.ObjViewCmdLineParser.{collectBool, collectVec}
import de.sciss.mellite.{MessageException, ObjView}
import de.sciss.processor.Processor.Aborted
import org.rogach.scallop.{ScallopConf, ValueConverter}
import org.rogach.scallop.exceptions.{Help, ScallopException, ScallopResult, Version}

import scala.collection.immutable.{IndexedSeq => Vec, Seq => ISeq}
import scala.concurrent.Future
import scala.util.control.NonFatal
import scala.util.{Failure, Success, Try}

object ObjViewCmdLineParser {
  def apply[T <: Txn[T]](f: ObjView.Factory, args: ISeq[String]): ObjViewCmdLineParser[f.Config[T]] =
    new ObjViewCmdLineParser[f.Config[T]](f, args)

  val collectBool: PartialFunction[String, Boolean] = {
    case "0" | "F" | "false"  => false
    case "1" | "T" | "true"   => true
  }

  def collectVec[A](s: String)(implicit peer: ValueConverter[A]): Option[Vec[A]] =
    s.split(',').filter(_.nonEmpty).foldLeft(Option(Vec.empty[A])) {
      case (Some(v0), w) =>
        peer.parse(("", w :: Nil) :: Nil) match {
          case Left(_)    => None
          case Right(opt) => opt.map(v0 :+ _)
        }
      case (None, _) => None
    }
}
class ObjViewCmdLineParser[C](private val f: ObjView.Factory, args: ISeq[String])
  extends ScallopConf(args) {

  import org.rogach.scallop._

  printedName = f.prefix
  banner(printedName)
  footer("")

  private[this] var _error    = Option.empty[String]
  private[this] var _aborted  = false

  override protected def onError(e: Throwable): Unit = e match {
    case r: ScallopResult if !throwError.value => r match {
      case Help("") =>
        builder.printHelp()
        _aborted = true
      case Help(subName) =>
        builder.findSubbuilder(subName).get.printHelp()
        _aborted = true
      case Version =>
        builder.vers.foreach(println)
        _aborted = true
      case ScallopException(message) =>
        _error = Some(message)
      case _ =>
        _error = Option(e.getMessage)
    }
    case _ => throw e
  }

  type Opt[A] = ScallopOption[A]

  val name: Opt[String] = opt(descr = s"Object's name (default: ${f.prefix})", default = Some(f.prefix))

  def nameOption: Option[String] = {
    val n = name()
    if (n.isEmpty) None else Some(n)
  }

  def boolArg(name    : String  = null,
              descr   : String  = "",
              required: Boolean = true
             ): ScallopOption[Boolean] =
    trailArg[String](name = name, descr = descr, required = required,
      validate = collectBool.isDefinedAt
    ).collect(collectBool)

  def boolOpt(name    : String  = null,
              descr   : String  = "",
              required: Boolean = true,
              default : Option[Boolean] = Some(false)
             ): ScallopOption[Boolean] =
    opt[String](name = name, descr = descr, required = required,
      validate = collectBool.isDefinedAt
    ).collect(collectBool).orElse(default)

  def nameOr(fallback: => Option[String]): String = {
    val n = name()
    if (n.isEmpty) fallback.getOrElse(n) else n
  }

  def vecArg[A](name    : String  = null,
                descr   : String  = "",
                required: Boolean = true)(implicit peer: ValueConverter[A]): ScallopOption[Vec[A]] =
    trailArg[String](name = name, descr = descr, required = required,
      validate = collectVec[A](_).isDefined
    ).map(collectVec[A](_).get) // XXX TODO d'oh this is ugly

  def parse(ok: => C): Try[C] = {
    verify()
    if (_aborted) Failure(Aborted()) else _error match {
      case Some(msg)  => Failure(MessageException(msg))
      case None       => Success(ok)
    }
  }

  def parseFut(ok: => Future[C]): Future[C] =
    try {
      verify()
      if (_aborted) Future.failed(Aborted()) else _error match {
        case Some(msg)  => Future.failed(MessageException(msg))
        case None       => ok
      }
    } catch {
      case NonFatal(ex) => Future.failed(ex)
    }
}